/***************************************************************************
                          kwth.cpp  -  description
                             -------------------
    begin                : Wed May 29 13:26:07 EDT 2002
    copyright            : (C) 2002 by woody
    email                : woody@xandros.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#define $VERSION "0.18"
#define WRITE_SYSLOG 1	/* store new metar lines in /var/log/messages */

#include <kglobal.h>
#include <klocale.h>
#include <kconfig.h>
#include <kapp.h>
#include <kmessagebox.h>
#include <qtimer.h>
#include <qpainter.h>
#include <qlineedit.h>
#include <qevent.h>
#include <qlabel.h>
#include <kstddirs.h>
#include <stdio.h>
#include <fcntl.h>

#include "kwth.h"
#include "kwthconfig.h"

/* get metar data section */
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <getopt.h>
#include <syslog.h>

#include <time.h>
#include <sys/time.h>
#include <pwd.h>
#include <sys/types.h>

int make_connect(const char* host);
int get_metar_html(char* buf_in, int buf_size, const char* metar_host,
		const char* metar_path, const char* site);
int get_metar(char* metar, int metarsize, const char* metar_host,
		const char* metar_path, const char* site);
int metar(int* metar_temperature, int* metar_dewpoint, int* metar_windspeed, int* metar_winddir);

/* Flag set by `--debug'. */
static int debug_flag = 0;

#define CONFIG_FILE	".kde/share/config/kwthrc"

#define	BAD_READING -999
#define METARSIZE 256
int	day;
int	hour;
int	minutes;
int	myzone = -4;
int	temperature = BAD_READING;
int	dew_point = BAD_READING;
int	wind_speed = BAD_READING;
int	wind_dir = BAD_READING;
int	old_temperature = BAD_READING;
int	old_dew_point = BAD_READING;
int	old_wind_speed = BAD_READING;
int	old_wind_dir = BAD_READING;
char	hint_text[128];
char	old_metar[METARSIZE];

char	confMetar[8] = {"CYOW"};;
int	confFrequency = 10;
char config_file[256] = {"~/.kde/share/config/kwthrc"};

QTimer	*runtimer;

int make_connect(const char* host)
{
struct hostent *ptrh;
struct protoent *ptrp;
struct sockaddr_in sad;
int sd;
int port = 80;

	memset((char *)&sad,0,sizeof(sad));
	sad.sin_family = AF_INET;

	sad.sin_port = htons((u_short)port);

	ptrh = gethostbyname(host);

	if ( ((char *)ptrh) == NULL )
	{
		fprintf(stderr,"invalid host: %s\n", host);
		return(-1);
	}

	memcpy(&sad.sin_addr, ptrh->h_addr, ptrh->h_length);

	if ( ((int)(ptrp = getprotobyname("tcp"))) == 0 )
	{
		fprintf(stderr,"cannot map \"tcp\" to protocol number\n");
		return(-1);
	}

	sd = socket(PF_INET, SOCK_STREAM, ptrp->p_proto);

	if (sd < 0)
	{
		fprintf(stderr, "Socket creation failed\n");
		return(-1);
	}

	if (connect(sd, (struct sockaddr *)&sad, sizeof(sad)) < 0)
	{
		fprintf(stderr, "Connect failed\n");
		close (sd);
		return(-1);
	}

	return(sd);
}

int get_metar_html(char* buf_in, int buf_size,
	const char* metar_host, const char* metar_path, const char* site)
{
char	buf_out[512];
int	response_size;
char	*base_get = "GET ";
int	sd;
int	result;

	sd = make_connect(metar_host);	/* "www.rap.ucar.edu" */
	if(sd < 0)
	{
		return (sd);
	}

	strcpy(buf_out, base_get);
	strcat(buf_out, metar_path);
	strcat(buf_out, " HTTP/1.0\n\n");
	if(debug_flag)
	{
		printf("%s\n", buf_out);
	}

	result = send(sd, buf_out, strlen(buf_out), 0);
	if(result < 0)
	{
		close(sd);
		return (result);
	}

	/* read the response back. start off with a NUL */
	response_size = 0;

	/* wait at most 60 seconds for the server to respond */
//	alarm(60);
	do
	{

		result = recv(sd, buf_in + response_size, buf_size - response_size - 1, 0);

		if (result > 0)
		{
			/* append the part of the page read */
			response_size += result;

			if (strstr(buf_in, "</HTML>"))
				break;
		}
	} while ((result > 0) && (response_size < buf_size - 1));

	buf_in[response_size] = '\0';

	if (debug_flag)
	{
		printf("%s\n", buf_in);
	}

	if ((result > 0) && !strstr(buf_in, " 200 OK"))
	{
		result = -1;
	}

	if (debug_flag)
	{
		printf("Retrieved: %s\n", (result <= 0) ? "Failed" : "OK");
	}

	close(sd);
	return result;
}

int get_metar(char* metar,
		int metarsize,
		const char* metar_host,
		const char* metar_path,
		const char* site)
{
char buf_in[2*1024];
char	*p1, *p;
int	i;

	i = get_metar_html(buf_in, sizeof(buf_in), metar_host, metar_path, site);
	if (i < 0)
		return (-1);

	/* if Sorry returned in HTML, we have a bad metar code */
	p1 = strstr(buf_in, "Sorry,");
	if (p1)
		return (-2);

	/* filter out the HTML overhead, find line starting with "site" kword */
	p1 = strstr(buf_in, site);
	if (!p1)
		return (-2);

	/* cut off at the end of this line */
	p = strchr(p1, '\n');
	if (p)
		*p = '\0';

	/* if some HTML code in this line (e.g. <br> - filter out */
	p = strchr(p1, '<');
	if (p)
		*p = '\0';

	strncpy(metar, p1, metarsize-2);
	metar[metarsize-1] = '\0';	/* make sure string is terminated */

	if (!strncmp(old_metar, metar, metarsize))
		return (-3);		/* string did not change */
	else
	{
		strncpy(old_metar, metar,
			metarsize<METARSIZE ? metarsize : METARSIZE);
		old_metar[METARSIZE - 1] = '\0';
	}
	return (0);
}

#define MAX_METAR_TOKENS	16

int	tokenize_metar(const char* metar, void* token_array)
{
char*	p;
int	token = 0;
char**	tok = (char**)token_array;

	/* if empty string - return 0 tokens */
	if (*metar == 0)
	{
		*tok = 0;
		return(0);
	}

	/* we have at least 1 token = the string */
	*tok++ = (char*)metar;
	token++;

	if ((p = strchr(metar, ' ')) != 0)
	{
		do
		{
			*p = 0;		/* convert word to string */
			p++;		/* skip to start of next word */
			*tok++ = p;	/* save as start of next word */

			if (token++ == MAX_METAR_TOKENS) break;
			p = strchr(p,' '); /* is it a word? */
		} while (p);
	}
	return(token);
}


//Example metar html URL for Ottawa
// This cgi-bin works for Nort America
//http://www.rap.ucar.edu/cgi-bin/rtwx_metars?metar_ids=CYOW
//#define METAR_HOST "www.rap.ucar.edu"
//#define METAR_PATH "/cgi-bin/rtwx_metars?metar_ids="
//#define SITE_CODE "CYOW"

// This php path works for the world
//http://www.rap.ucar.edu/weather/surface/index.php?metarIds=CYOW&hoursStr=most+recent+only&std_trans=standard&num_metars=number
#define METAR_HOST "www.rap.ucar.edu"
#define METAR_PATH_BEG "/weather/surface/index.php?metarIds="
#define METAR_PATH_END "&hoursStr=most+recent+only&std_trans=standard&num_metars=number"

int metar(int* metar_temperature, int* metar_dewpoint, int* metar_windspeed, int* metar_winddir)
{
char	metar[METARSIZE];
char	metar_words[METARSIZE];
char*	metar_tokens[MAX_METAR_TOKENS];
int	tokens;
int	i;
char	metar_path[512];

#if 1
	debug_flag = 0;

	strcpy(metar_path, METAR_PATH_BEG); /* "/weather/surface/index.php?metarIds=" */
	strcat(metar_path, confMetar);	    /* "CYOW" */
	strcat(metar_path, METAR_PATH_END); /* "&hoursStr=most+recent+only&std_trans=standard&num_metars=number" */

	i = get_metar(metar, 256, METAR_HOST, metar_path, confMetar);
	if (i)
		return -1;
	if (debug_flag)
	{
		printf("%s\n", metar);
	}
	/*  CYOW 270100Z 28005KT 15SM FEW070 04/M07 A3021 RMK AC1 SLP235= */
#else
	strcpy(metar,"CYOW 270100Z 28005KT 15SM FEW070 14/07 A3021 RMK AC1 SLP235=");
#endif
#ifdef WRITE_SYSLOG
	openlog("Kwth", 0, LOG_USER);
	syslog(LOG_INFO, metar);
	closelog();
#endif
	strncpy (metar_words,metar, sizeof(metar_words)-2);
	metar_words[sizeof(metar_words)-1] = 0;

	tokens = tokenize_metar(metar_words, &metar_tokens);
	if (tokens < 2)
		return -2;

        if (debug_flag)
	{
		for(i = 0; i<tokens; i++)
		{
			printf("%d: %s\n", i, metar_tokens[i]);
		}
	}

	/* extract time info */
	for (i = 0; i<tokens; i++)
	{
	char*	p;
	int	j;

		p = metar_tokens[i];
		j = strlen(p);
		if (*(p+j-1) == 'Z')
		{
			day = (*p++ - '0')*10;
			day += (*p++ - '0');
			hour = (*p++ - '0')*10;
			hour += (*p++ - '0');
			minutes = (*p++ - '0')*10;
			minutes += (*p++ - '0');
			hour += myzone;
			if (hour >= 24) hour -= 24;
			if (hour < 0) hour += 24;
			break;
		}
	}

	/* extract temperature */
	*metar_temperature = BAD_READING;
	*metar_dewpoint = BAD_READING;

	for (i = 0; i<tokens; i++)
	{
	char*	p;
	int	temperature;
	int	dew_temp;
	int     j;

		if ((p = strchr(metar_tokens[i],'/')) != 0)
		{
			/* make sure tokens does end with digit */
			j = strlen(metar_tokens[i]);
			j = *(metar_tokens[i] + j - 1);
			if (j > '9' || j < '0') continue;

			*p = 0;
			if (*metar_tokens[i] == 'M')
				*metar_tokens[i] = '-';
			p++;
			if (*p == 'M')
				*p = '-';
			temperature = atoi(metar_tokens[i]);
			dew_temp = atoi(p);
//			printf("temperature = %d, dew point = %d.\n",temperature, dew_temp);
			*metar_temperature = temperature;
			*metar_dewpoint = dew_temp;
			break;
		}
	}

	/* extract wind data - look for word ending with KT */
	*metar_windspeed = BAD_READING;
	*metar_winddir = BAD_READING;

	for (i = 0; i<tokens; i++)
	{
	char*	p;
	int	wind_speed;
	int	wind_dir;
	int	j;

		p = metar_tokens[i];
		j = strlen(p);
		if ((*(p+j-2) == 'K') && (*(p+j-1) == 'T'))
		{
			*(p+j-2) = 0;
			wind_dir = (*p++ - '0')*100;
			wind_dir += (*p++ - '0')*10;
			wind_dir += (*p++ - '0');
			wind_speed = atoi(p);
//			printf("wind dir = %d, speed = %d.\n",wind_dir, (int)(0.5+wind_speed*1.82));
			/* convert knots to km/h, rounding up/down */
			*metar_windspeed = (int)(0.5+wind_speed*1.82);
			*metar_winddir = wind_dir;
			break;
		}
	}

	/* extract another form of wind data - look for word ending with MPS */
	/* EPKK 311330Z VRB02MPS 9999 FEW050 SCT100 18/09 Q1022 745.8 */
	/* EPWA 311300Z 35005MPS 320V050 CAVOK 20/07 Q1021 NOSIG */
	if (*metar_windspeed == BAD_READING && *metar_winddir == BAD_READING)
	{
	for (i = 0; i<tokens; i++)
	{
	char*	p;
	int	wind_speed;
	int	wind_dir;
	int	j;
	char	wdir[4];

		p = metar_tokens[i];
		j = strlen(p);
		if ((*(p+j-3) == 'M') && (*(p+j-2) == 'P') && (*(p+j-1) == 'S'))
		{
			*(p+j-3) = 0;
			wdir[0] = *p++;
			wdir[1] = *p++;
			wdir[2] = *p++;
			wdir[3] = 0;
			wind_dir = atoi(wdir);
#if 0
			if (
			j = strtol(&wind_dir, (char **)NULL, 10);

			}
			else
			{
			wind_dir = BAD_READING;
			p += 3;		/* skip the garbage */
			}
#endif
			wind_speed = atoi(p);

//			printf("wind dir = %d, speed = %d.\n",wind_dir, (int)(0.5+wind_speed*1.82));
			/* convert m/s to km/h, rounding up/down */
			*metar_windspeed = (int)(0.5+wind_speed*3.6);
			*metar_winddir = wind_dir;
			break;
		}
	}
	}
	return (0);
}
/* end of get metar data section */

extern "C"
{
KPanelApplet* init( QWidget *parent, const QString configFile)
{
	KGlobal::locale()->insertCatalogue("kwth");
	return new Kwth(configFile, KPanelApplet::Normal,
		KPanelApplet::About | KPanelApplet::Help | KPanelApplet::Preferences,
		parent, "kwth");
}
}

Kwth::Kwth(const QString& configFile, Type type, int actions, QWidget *parent, const char *name)
       : KPanelApplet(configFile, type, actions, parent, name)
{
struct	timeval	ttime;
struct	timezone tzone;
struct	passwd*	userdata;

	userdata = getpwuid(getuid());
	sprintf(config_file,"%s/%s",userdata->pw_dir,CONFIG_FILE);

	gettimeofday(&ttime, &tzone);
	/* tz_minuteswest does not work on Netwinders - libc bug? */
#if 1
	/* force update of global timezone and daylight vars */
	localtime(&ttime.tv_sec);
	myzone = (daylight - timezone)/3600;
#else
	myzone = -tzone.tz_minuteswest/60;
#endif
	// Get the current application configuration handle
	ksConfig = config();

	KStandardDirs& dirs = *(kapp->dirs()); // Get application directories

	// Setup a timer until update is needed...
	timer = new QTimer( this );
	connect( timer, SIGNAL(timeout()),this, SLOT(notify()) );
	runtimer = timer;

	// Setup this timer to call our notify() method in every 10 minutes...
	// 60*1000 = 1 minute
	timer->start( 60*1000*confFrequency , FALSE);

	// load pixmaps
//	cloud_pixmap = new QPixmap(dirs.findResource("data","kwth/pixmaps/charge.png"));

//	my_hint = new QString;

//#ifdef WRITE_SYSLOG
//	openlog("Kwth", 0, LOG_USER);
//#endif
	notify();
}


/* left button - display message box */
/* right button - update the reading */
/* middle button - exit */

void Kwth::mouseReleaseEvent(QMouseEvent * e)
{
	if (e->button() == LeftButton)
	{
		char	boxtext[2*METARSIZE];
		char	boxtitle[64];

		sprintf(boxtext,"%s\n%s",hint_text,old_metar);
		sprintf(boxtitle,"Kwth %s: Current weather for %s",$VERSION, confMetar);
		KMessageBox::information(this,i18n(boxtext),i18n(boxtitle));
	}

	if (e->button() == MidButton)		/* or MidButton */
	{
		preferences();
	}

	if (e->button() == RightButton)		/* or MidButton */
	{
		preferences();
	}
}

/** The method invoked by the timer */
void Kwth::notify()
{

	if (temperature != BAD_READING)
		old_temperature = temperature;
	if (dew_point != BAD_READING)
		old_dew_point = dew_point;
	if (wind_speed != BAD_READING)
		old_wind_speed = wind_speed;
	if (wind_dir != BAD_READING)
		old_wind_dir = wind_dir;

	metar(&temperature, &dew_point, &wind_speed, &wind_dir);
	// Finally call the display method to show the wheather info.
	display_kwth();
}

void Kwth::display_kwth()
{
QString	hint;
QColor	color;

char	wri [12];
char	wdir [4];

int	cur_temperature;
int	cur_dew_point;
int	cur_wind_speed;
int	cur_wind_dir;

/** Display the information got from the APM here... */
	int ysh = (height()-42)/2;


	// First clear the display and create a painter
	erase();
	QPainter *p = new QPainter (this);

	// Now print the information bar below
	QFont font("helvetica",8);
	p->setFont(font);

	/* setup current temperature */
	if (temperature != BAD_READING)
	{
		cur_temperature = temperature;
		color = blue;
	}
	else if (old_temperature != BAD_READING)
	{
		cur_temperature = old_temperature;
		color = black;
	}
	else
	{
		cur_temperature = BAD_READING;
		color = red;
	}

	if (cur_temperature != BAD_READING)
		sprintf(wri,"%dC",cur_temperature);
	else
		sprintf(wri,"??C");


	/* display temperature */
	QBrush brush(white);		//(color);
//	p->fillRect(0,ysh+10,1,20,brush);	//vert bar on left border
//	p->fillRect(0,ysh+10,48,1,brush);	//horiz bar above temp
//	p->fillRect(0,ysh+21,36,1,brush);	//horiz bar in the middle
	p->fillRect(0,ysh+10,36,22,brush);	//paint rect


	p->setPen(color);
	p->drawText(1,ysh+20,wri);


	/* setup current dew point */
	if (dew_point != BAD_READING)
		cur_dew_point = dew_point;
	else if (old_dew_point != BAD_READING)
		cur_dew_point = old_dew_point;
	else
		cur_dew_point = BAD_READING;

	/* setup current wind direction */
	if (wind_dir != BAD_READING)
		cur_wind_dir = wind_dir;
	else if (old_wind_dir != BAD_READING)
		cur_wind_dir = old_wind_dir;
	else
		cur_wind_dir = BAD_READING;

	if (cur_wind_dir != BAD_READING)
	{
		if (cur_wind_dir > 360)
        		cur_wind_dir %= 360;

		if (cur_wind_dir < 13) strcpy (wdir, "N");
		else if (cur_wind_dir < 45-12) strcpy (wdir, "NNE");
		else if (cur_wind_dir < 45+12) strcpy (wdir, "NE");
		else if (cur_wind_dir < 90-12) strcpy (wdir, "ENE");
		else if (cur_wind_dir < 90+12) strcpy (wdir, "E");
		else if (cur_wind_dir < 135-12) strcpy (wdir, "ESE");
		else if (cur_wind_dir < 135+12) strcpy (wdir, "SE");
		else if (cur_wind_dir < 180-12) strcpy (wdir, "SSE");
		else if (cur_wind_dir < 180+12) strcpy (wdir, "S");
		else if (cur_wind_dir < 225-12) strcpy (wdir, "SSW");
		else if (cur_wind_dir < 225+12) strcpy (wdir, "SW");
		else if (cur_wind_dir < 270-12) strcpy (wdir, "WSW");
		else if (cur_wind_dir < 270+12) strcpy (wdir, "W");
		else if (cur_wind_dir < 315-12) strcpy (wdir, "WNW");
		else if (cur_wind_dir < 315+12) strcpy (wdir, "NW");
		else if (cur_wind_dir < 360-12) strcpy (wdir, "NNW");
		else strcpy (wdir, "N");
	}
	else strcpy (wdir, "???");


	/* setup current wind speed */
	if (wind_speed != BAD_READING)
		cur_wind_speed = wind_speed;
	else if (old_wind_speed != BAD_READING)
		cur_wind_speed = old_wind_speed;
	else
		cur_wind_speed = BAD_READING;

	if (cur_wind_speed != BAD_READING)
	{
		if (!cur_wind_speed)
			sprintf(wri,"CALM");
		else
		{
			if (wdir[2] == 0)
				sprintf(wri,"%d %s",wind_speed, wdir);
			else
				sprintf(wri,"%d%s",wind_speed, wdir);
		}
	}
	else
		sprintf(wri,"???");


	/* display wind info */
//	QFont font1("helvetica",6);
//	p->setFont(font1);
//	p->setPen(color);
	p->drawText(1,ysh+30,wri);


	/* build hint line */
//	sprintf(hint_text,"Temperature %dC, dew point %dC, wind %d km/h",
//		temperature, dew_point, wind_speed);

	sprintf(hint_text,"Report time %02d:%02d, temperature ", hour, minutes);
	if (cur_temperature != BAD_READING)
	{
		sprintf(wri,"%d", cur_temperature);
		strcat(hint_text, wri);
	}
	else
		strcat(hint_text, "??");

	strcat (hint_text,"C, dew point ");
	if (cur_dew_point != BAD_READING)
	{
		sprintf(wri,"%d", cur_dew_point);
		strcat(hint_text, wri);
	}
	else
		strcat(hint_text, "??");

	strcat (hint_text,"C, wind ");
	if (cur_wind_speed != BAD_READING)
	{
		sprintf(wri,"%d", cur_wind_speed);
		strcat(hint_text, wri);
	}
	else
		strcat(hint_text, "??");
	strcat (hint_text, " km/h");

	if ((cur_wind_speed != BAD_READING) && cur_wind_speed)
	{
		strcat(hint_text," from ");
		strcat(hint_text,wdir);
	}
	hint.append(hint_text);

	p->end();
	delete p;
}

Kwth::~Kwth()
{
	runtimer->stop();
}


void Kwth::preferences()
{
char confFreq[8];
char tempMetar[8];
int	i;
int	changeFlag = 0;

	sprintf(confFreq,"%d", confFrequency);

	kwthConfig dlg(this, "Config", true);

	dlg.TitleLabel->setText("Kwth "$VERSION" by Woody Suwalski\nwoody@xandros.com");
	dlg.MetarCode->setText(confMetar);
	dlg.Frequency->setText(confFreq);
	dlg.exec();
	i = dlg.result();

	// result 1 == OK, 0 = Cancel
	if (i == 1)
	{
		strncpy (confFreq, dlg.Frequency->text(), 4);
		confFreq[4] = 0;
		i = atoi(confFreq);
		if ((confFrequency != i) && (i > 0))
		{
			confFrequency = atoi(confFreq);
			changeFlag |= 1;
			runtimer->changeInterval(60*1000*confFrequency);
		}

		strcpy (tempMetar, dlg.MetarCode->text());
		if (strcmp(tempMetar, confMetar))
		{
			strcpy (confMetar, tempMetar);
			changeFlag |= 2;
			notify();
		}
		if (changeFlag)
			writeConfig();
	}
}

int Kwth::readConfig( void)
{
FILE*	f = NULL;
char	buf[64];
char	*val;
char	temp[16];
int		i;

	f = fopen(config_file, "r");

	if (!f)
	{
//		printf("Read error: %s: %d (%s).\n",
//			config_file, errno, strerror(errno));
		return (writeConfig());
	}
	else
	{
		while (fgets (buf, sizeof(buf)-1, f))
		{
			if ((val = strchr(buf,'='))!=NULL)
			{
				*val = 0;
				val++;

				if (strncmp (buf, "metar", 5) == 0)
				{
					i = sscanf(val, "%s", temp);
					if(i == 1)
						strncpy(confMetar, temp, 4);
				}
				else if (strncmp (buf, "frequency", 9) == 0)
				{
					i = atoi (val);
					if ((i > 0) && (i < 600))
						confFrequency = i;
				}
			}
		}
		fclose (f);
	}
	return 0;
}


int Kwth::writeConfig( void)
{
FILE*	f = NULL;
char	buf[64];

	f = fopen(config_file, "w");
	if (!f)
	{
//		printf("Write error: %s: %d (%s).\n",
//			config_file, errno, strerror(errno));
		return 1;
	}
	sprintf (buf,"[ kwth "$VERSION" config file ]\n\n");
	fputs(buf, f);

	sprintf (buf,"metar = %s\n", confMetar);
	fputs(buf, f);

	sprintf (buf,"frequency = %d\n", confFrequency);
	fputs(buf, f);

	fclose(f);

	return 0;
}

int Kwth::widthForHeight( int height ) const
{
	return 36;
}

int Kwth::heightForWidth( int width ) const
{
	return 24;
}

void Kwth::resizeEvent( QResizeEvent *_Event)
{

}

void Kwth::paintEvent ( QPaintEvent * e)
{
	display_kwth();
}
